@@ -9,11 +9,10 @@
 
 use regex::Regex;
 use std::fmt;
+use tree_sitter::CaptureQuantifier;
 use tree_sitter::Language;
 use tree_sitter::Query;
 
-use crate::Context;
-use crate::DisplayWithContext;
 use crate::Identifier;
 use crate::Location;
 
@@ -21,6 +20,8 @@ use crate::Location;
 #[derive(Debug)]
 pub struct File {
     pub language: Language,
+    /// The combined query of all stanzas in the file
+    pub query: Option<Query>,
     /// The list of stanzas in the file
     pub stanzas: Vec<Stanza>,
 }
@@ -29,6 +30,7 @@ impl File {
     pub fn new(language: Language) -> File {
         File {
             language,
+            query: None,
             stanzas: Vec::new(),
         }
     }
@@ -41,6 +43,8 @@ pub struct Stanza {
     pub query: Query,
     /// The list of statements in the stanza
     pub statements: Vec<Statement>,
+    /// Capture index of the full match in the file query
+    pub full_match_file_capture_index: usize,
     pub location: Location,
 }
 
@@ -61,20 +65,26 @@ pub enum Statement {
     Scan(Scan),
     // Debugging
     Print(Print),
+    // If
+    If(If),
+    // ForIn
+    ForIn(ForIn),
 }
 
-impl DisplayWithContext for Statement {
-    fn fmt(&self, f: &mut fmt::Formatter, ctx: &Context) -> fmt::Result {
+impl std::fmt::Display for Statement {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         match self {
-            Statement::DeclareImmutable(stmt) => stmt.fmt(f, ctx),
-            Statement::DeclareMutable(stmt) => stmt.fmt(f, ctx),
-            Statement::Assign(stmt) => stmt.fmt(f, ctx),
-            Statement::CreateGraphNode(stmt) => stmt.fmt(f, ctx),
-            Statement::AddGraphNodeAttribute(stmt) => stmt.fmt(f, ctx),
-            Statement::CreateEdge(stmt) => stmt.fmt(f, ctx),
-            Statement::AddEdgeAttribute(stmt) => stmt.fmt(f, ctx),
-            Statement::Scan(stmt) => stmt.fmt(f, ctx),
-            Statement::Print(stmt) => stmt.fmt(f, ctx),
+            Self::DeclareImmutable(stmt) => stmt.fmt(f),
+            Self::DeclareMutable(stmt) => stmt.fmt(f),
+            Self::Assign(stmt) => stmt.fmt(f),
+            Self::CreateGraphNode(stmt) => stmt.fmt(f),
+            Self::AddGraphNodeAttribute(stmt) => stmt.fmt(f),
+            Self::CreateEdge(stmt) => stmt.fmt(f),
+            Self::AddEdgeAttribute(stmt) => stmt.fmt(f),
+            Self::Scan(stmt) => stmt.fmt(f),
+            Self::Print(stmt) => stmt.fmt(f),
+            Self::If(stmt) => stmt.fmt(f),
+            Self::ForIn(stmt) => stmt.fmt(f),
         }
     }
 }
@@ -94,16 +104,11 @@ impl From<AddEdgeAttribute> for Statement {
     }
 }
 
-impl DisplayWithContext for AddEdgeAttribute {
-    fn fmt(&self, f: &mut fmt::Formatter, ctx: &Context) -> fmt::Result {
-        write!(
-            f,
-            "attr ({} -> {})",
-            self.source.display_with(ctx),
-            self.sink.display_with(ctx),
-        )?;
+impl std::fmt::Display for AddEdgeAttribute {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        write!(f, "attr ({} -> {})", self.source, self.sink)?;
         for attr in &self.attributes {
-            write!(f, " {}", attr.display_with(ctx))?;
+            write!(f, " {}", attr)?;
         }
         write!(f, " at {}", self.location)
     }
@@ -123,11 +128,11 @@ impl From<AddGraphNodeAttribute> for Statement {
     }
 }
 
-impl DisplayWithContext for AddGraphNodeAttribute {
-    fn fmt(&self, f: &mut fmt::Formatter, ctx: &Context) -> fmt::Result {
-        write!(f, "attr ({})", self.node.display_with(ctx),)?;
+impl std::fmt::Display for AddGraphNodeAttribute {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        write!(f, "attr ({})", self.node)?;
         for attr in &self.attributes {
-            write!(f, " {}", attr.display_with(ctx),)?;
+            write!(f, " {}", attr)?;
         }
         write!(f, " at {}", self.location)
     }
@@ -147,14 +152,12 @@ impl From<Assign> for Statement {
     }
 }
 
-impl DisplayWithContext for Assign {
-    fn fmt(&self, f: &mut fmt::Formatter, ctx: &Context) -> fmt::Result {
+impl std::fmt::Display for Assign {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         write!(
             f,
             "set {} = {} at {}",
-            self.variable.display_with(ctx),
-            self.value.display_with(ctx),
-            self.location,
+            self.variable, self.value, self.location,
         )
     }
 }
@@ -166,14 +169,9 @@ pub struct Attribute {
     pub value: Expression,
 }
 
-impl DisplayWithContext for Attribute {
-    fn fmt(&self, f: &mut fmt::Formatter, ctx: &Context) -> fmt::Result {
-        write!(
-            f,
-            "{} = {}",
-            self.name.display_with(ctx),
-            self.value.display_with(ctx),
-        )
+impl std::fmt::Display for Attribute {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        write!(f, "{} = {}", self.name, self.value)
     }
 }
 
@@ -191,14 +189,12 @@ impl From<CreateEdge> for Statement {
     }
 }
 
-impl DisplayWithContext for CreateEdge {
-    fn fmt(&self, f: &mut fmt::Formatter, ctx: &Context) -> fmt::Result {
+impl std::fmt::Display for CreateEdge {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         write!(
             f,
             "edge {} -> {} at {}",
-            self.source.display_with(ctx),
-            self.sink.display_with(ctx),
-            self.location,
+            self.source, self.sink, self.location,
         )
     }
 }
@@ -216,14 +212,9 @@ impl From<CreateGraphNode> for Statement {
     }
 }
 
-impl DisplayWithContext for CreateGraphNode {
-    fn fmt(&self, f: &mut fmt::Formatter, ctx: &Context) -> fmt::Result {
-        write!(
-            f,
-            "node {} at {}",
-            self.node.display_with(ctx),
-            self.location
-        )
+impl std::fmt::Display for CreateGraphNode {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        write!(f, "node {} at {}", self.node, self.location)
     }
 }
 
@@ -241,14 +232,12 @@ impl From<DeclareImmutable> for Statement {
     }
 }
 
-impl DisplayWithContext for DeclareImmutable {
-    fn fmt(&self, f: &mut fmt::Formatter, ctx: &Context) -> fmt::Result {
+impl std::fmt::Display for DeclareImmutable {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         write!(
             f,
             "let {} = {} at {}",
-            self.variable.display_with(ctx),
-            self.value.display_with(ctx),
-            self.location,
+            self.variable, self.value, self.location,
         )
     }
 }
@@ -267,14 +256,12 @@ impl From<DeclareMutable> for Statement {
     }
 }
 
-impl DisplayWithContext for DeclareMutable {
-    fn fmt(&self, f: &mut fmt::Formatter, ctx: &Context) -> fmt::Result {
+impl std::fmt::Display for DeclareMutable {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         write!(
             f,
             "var {} = {} at {}",
-            self.variable.display_with(ctx),
-            self.value.display_with(ctx),
-            self.location,
+            self.variable, self.value, self.location,
         )
     }
 }
@@ -292,11 +279,11 @@ impl From<Print> for Statement {
     }
 }
 
-impl DisplayWithContext for Print {
-    fn fmt(&self, f: &mut fmt::Formatter, ctx: &Context) -> fmt::Result {
+impl std::fmt::Display for Print {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         write!(f, "print")?;
         for val in &self.values {
-            write!(f, " {},", val.display_with(ctx),)?;
+            write!(f, " {},", val)?;
         }
         write!(f, " at {}", self.location)
     }
@@ -316,14 +303,9 @@ impl From<Scan> for Statement {
     }
 }
 
-impl DisplayWithContext for Scan {
-    fn fmt(&self, f: &mut fmt::Formatter, ctx: &Context) -> fmt::Result {
-        write!(
-            f,
-            "scan {} {{ ... }} at {}",
-            self.value.display_with(ctx),
-            self.location
-        )
+impl std::fmt::Display for Scan {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        write!(f, "scan {} {{ ... }} at {}", self.value, self.location)
     }
 }
 
@@ -343,12 +325,126 @@ impl PartialEq for ScanArm {
     }
 }
 
-impl DisplayWithContext for ScanArm {
-    fn fmt(&self, f: &mut fmt::Formatter, _ctx: &Context) -> fmt::Result {
+impl std::fmt::Display for ScanArm {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         write!(f, "{:?} {{ ... }}", self.regex.as_str())
     }
 }
 
+/// A `cond` conditional statement that selects the first branch with a matching condition
+#[derive(Debug, Eq, PartialEq)]
+pub struct If {
+    pub arms: Vec<IfArm>,
+    pub location: Location,
+}
+
+impl From<If> for Statement {
+    fn from(statement: If) -> Statement {
+        Statement::If(statement)
+    }
+}
+
+impl std::fmt::Display for If {
+    fn fmt(&self, f: &mut fmt::Formatter) -> std::fmt::Result {
+        let mut first = true;
+        for arm in &self.arms {
+            if first {
+                first = false;
+                write!(f, "if {} {{ ... }}", DisplayConditions(&arm.conditions))?;
+            } else {
+                if !arm.conditions.is_empty() {
+                    write!(f, " elif {} {{ ... }}", DisplayConditions(&arm.conditions))?;
+                } else {
+                    write!(f, " else {{ ... }}")?;
+                }
+            }
+        }
+        write!(f, " at {}", self.location)
+    }
+}
+
+/// One arm of a `cond` statement
+#[derive(Debug, PartialEq, Eq)]
+pub struct IfArm {
+    pub conditions: Vec<Condition>,
+    pub statements: Vec<Statement>,
+    pub location: Location,
+}
+
+struct DisplayConditions<'a>(&'a Vec<Condition>);
+
+#[derive(Debug, PartialEq, Eq)]
+pub enum Condition {
+    Some {
+        value: Expression,
+        location: Location,
+    },
+    None {
+        value: Expression,
+        location: Location,
+    },
+    Bool {
+        value: Expression,
+        location: Location,
+    },
+}
+
+impl std::fmt::Display for DisplayConditions<'_> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> std::fmt::Result {
+        let mut first = true;
+        for condition in self.0.iter() {
+            if first {
+                first = false;
+                write!(f, "{}", condition)?;
+            } else {
+                write!(f, ", {}", condition)?;
+            }
+        }
+        Ok(())
+    }
+}
+
+impl std::fmt::Display for Condition {
+    fn fmt(&self, f: &mut fmt::Formatter) -> std::fmt::Result {
+        match self {
+            Condition::Some { value, .. } => {
+                write!(f, "some {}", value)
+            }
+            Condition::None { value, .. } => {
+                write!(f, "none {}", value)
+            }
+            Condition::Bool { value, .. } => {
+                write!(f, "{}", value)
+            }
+        }
+    }
+}
+
+/// A `for in` statement
+#[derive(Debug, Eq, PartialEq)]
+pub struct ForIn {
+    pub variable: UnscopedVariable,
+    pub value: Expression,
+    pub statements: Vec<Statement>,
+    pub location: Location,
+}
+
+impl From<ForIn> for Statement {
+    fn from(statement: ForIn) -> Statement {
+        Statement::ForIn(statement)
+    }
+}
+
+impl std::fmt::Display for ForIn {
+    fn fmt(&self, f: &mut fmt::Formatter) -> std::fmt::Result {
+        write!(
+            f,
+            "for {} in {} {{ ... }} at {}",
+            self.variable, self.value, self.location,
+        )
+    }
+}
+
 /// A reference to a variable
 #[derive(Debug, Eq, PartialEq)]
 pub enum Variable {
@@ -356,11 +452,11 @@ pub enum Variable {
     Unscoped(UnscopedVariable),
 }
 
-impl DisplayWithContext for Variable {
-    fn fmt(&self, f: &mut fmt::Formatter, ctx: &Context) -> fmt::Result {
+impl std::fmt::Display for Variable {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         match self {
-            Variable::Scoped(variable) => variable.fmt(f, ctx),
-            Variable::Unscoped(variable) => variable.fmt(f, ctx),
+            Variable::Scoped(variable) => variable.fmt(f),
+            Variable::Unscoped(variable) => variable.fmt(f),
         }
     }
 }
@@ -370,6 +466,7 @@ impl DisplayWithContext for Variable {
 pub struct ScopedVariable {
     pub scope: Box<Expression>,
     pub name: Identifier,
+    pub location: Location,
 }
 
 impl From<ScopedVariable> for Variable {
@@ -378,14 +475,9 @@ impl From<ScopedVariable> for Variable {
     }
 }
 
-impl DisplayWithContext for ScopedVariable {
-    fn fmt(&self, f: &mut fmt::Formatter, ctx: &Context) -> fmt::Result {
-        write!(
-            f,
-            "{}.{}",
-            self.scope.display_with(ctx),
-            self.name.display_with(ctx),
-        )
+impl std::fmt::Display for ScopedVariable {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        write!(f, "{}.{}", self.scope, self.name)
     }
 }
 
@@ -393,6 +485,7 @@ impl DisplayWithContext for ScopedVariable {
 #[derive(Debug, Eq, PartialEq)]
 pub struct UnscopedVariable {
     pub name: Identifier,
+    pub location: Location,
 }
 
 impl From<UnscopedVariable> for Variable {
@@ -401,15 +494,9 @@ impl From<UnscopedVariable> for Variable {
     }
 }
 
-impl From<Identifier> for Variable {
-    fn from(name: Identifier) -> Variable {
-        UnscopedVariable { name }.into()
-    }
-}
-
-impl DisplayWithContext for UnscopedVariable {
-    fn fmt(&self, f: &mut fmt::Formatter, ctx: &Context) -> fmt::Result {
-        write!(f, "{}", self.name.display_with(ctx))
+impl std::fmt::Display for UnscopedVariable {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        write!(f, "{}", self.name)
     }
 }
 
@@ -436,20 +523,20 @@ pub enum Expression {
     RegexCapture(RegexCapture),
 }
 
-impl DisplayWithContext for Expression {
-    fn fmt(&self, f: &mut fmt::Formatter, ctx: &Context) -> fmt::Result {
+impl std::fmt::Display for Expression {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         match self {
             Expression::FalseLiteral => write!(f, "false"),
             Expression::NullLiteral => write!(f, "#null"),
             Expression::TrueLiteral => write!(f, "true"),
-            Expression::IntegerConstant(expr) => expr.fmt(f, ctx),
-            Expression::StringConstant(expr) => expr.fmt(f, ctx),
-            Expression::List(expr) => expr.fmt(f, ctx),
-            Expression::Set(expr) => expr.fmt(f, ctx),
-            Expression::Capture(expr) => expr.fmt(f, ctx),
-            Expression::Variable(expr) => expr.fmt(f, ctx),
-            Expression::Call(expr) => expr.fmt(f, ctx),
-            Expression::RegexCapture(expr) => expr.fmt(f, ctx),
+            Expression::IntegerConstant(expr) => expr.fmt(f),
+            Expression::StringConstant(expr) => expr.fmt(f),
+            Expression::List(expr) => expr.fmt(f),
+            Expression::Set(expr) => expr.fmt(f),
+            Expression::Capture(expr) => expr.fmt(f),
+            Expression::Variable(expr) => expr.fmt(f),
+            Expression::Call(expr) => expr.fmt(f),
+            Expression::RegexCapture(expr) => expr.fmt(f),
         }
     }
 }
@@ -467,11 +554,11 @@ impl From<Call> for Expression {
     }
 }
 
-impl DisplayWithContext for Call {
-    fn fmt(&self, f: &mut fmt::Formatter, ctx: &Context) -> fmt::Result {
-        write!(f, "({}", self.function.display_with(ctx))?;
+impl std::fmt::Display for Call {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        write!(f, "({}", self.function)?;
         for arg in &self.parameters {
-            write!(f, " {}", arg.display_with(ctx))?;
+            write!(f, " {}", arg)?;
         }
         write!(f, ")")
     }
@@ -480,10 +567,15 @@ impl DisplayWithContext for Call {
 /// A capture expression that references a syntax node
 #[derive(Debug, Eq, PartialEq)]
 pub struct Capture {
-    /// The index of this capture in the block's tree-sitter query
-    pub index: usize,
     /// The name of the capture
     pub name: Identifier,
+    /// The suffix of the capture
+    pub quantifier: CaptureQuantifier,
+    /// Capture index in the merged file query
+    pub file_capture_index: usize,
+    /// Capture index in the stanza query
+    pub stanza_capture_index: usize,
+    pub location: Location,
 }
 
 impl From<Capture> for Expression {
@@ -492,9 +584,9 @@ impl From<Capture> for Expression {
     }
 }
 
-impl DisplayWithContext for Capture {
-    fn fmt(&self, f: &mut fmt::Formatter, ctx: &Context) -> fmt::Result {
-        write!(f, "{}", self.name.display_with(ctx))
+impl std::fmt::Display for Capture {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        write!(f, "@{}", self.name)
     }
 }
 
@@ -510,8 +602,8 @@ impl From<IntegerConstant> for Expression {
     }
 }
 
-impl DisplayWithContext for IntegerConstant {
-    fn fmt(&self, f: &mut fmt::Formatter, _ctx: &Context) -> fmt::Result {
+impl std::fmt::Display for IntegerConstant {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         write!(f, "{}", self.value)
     }
 }
@@ -528,16 +620,16 @@ impl From<ListComprehension> for Expression {
     }
 }
 
-impl DisplayWithContext for ListComprehension {
-    fn fmt(&self, f: &mut fmt::Formatter, ctx: &Context) -> fmt::Result {
+impl std::fmt::Display for ListComprehension {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         write!(f, "[")?;
         let mut first = true;
         for elem in &self.elements {
             if first {
-                write!(f, "{}", elem.display_with(ctx))?;
+                write!(f, "{}", elem)?;
                 first = false;
             } else {
-                write!(f, ", {}", elem.display_with(ctx))?;
+                write!(f, ", {}", elem)?;
             }
         }
         write!(f, "]")
@@ -556,8 +648,8 @@ impl From<RegexCapture> for Expression {
     }
 }
 
-impl DisplayWithContext for RegexCapture {
-    fn fmt(&self, f: &mut fmt::Formatter, _ctx: &Context) -> fmt::Result {
+impl std::fmt::Display for RegexCapture {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         write!(f, "${}", self.match_index)
     }
 }
@@ -574,16 +666,16 @@ impl From<SetComprehension> for Expression {
     }
 }
 
-impl DisplayWithContext for SetComprehension {
-    fn fmt(&self, f: &mut fmt::Formatter, ctx: &Context) -> fmt::Result {
+impl std::fmt::Display for SetComprehension {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         write!(f, "{{")?;
         let mut first = true;
         for elem in &self.elements {
             if first {
-                write!(f, "{}", elem.display_with(ctx))?;
+                write!(f, "{}", elem)?;
                 first = false;
             } else {
-                write!(f, ", {}", elem.display_with(ctx))?;
+                write!(f, ", {}", elem)?;
             }
         }
         write!(f, "}}")
@@ -602,8 +694,8 @@ impl From<StringConstant> for Expression {
     }
 }
 
-impl DisplayWithContext for StringConstant {
-    fn fmt(&self, f: &mut fmt::Formatter, _ctx: &Context) -> fmt::Result {
+impl std::fmt::Display for StringConstant {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         write!(f, "{:?}", self.value)
     }
 }
@@ -614,9 +706,9 @@ impl From<String> for Expression {
     }
 }
 
-impl From<Identifier> for Expression {
-    fn from(name: Identifier) -> Expression {
-        Expression::Variable(UnscopedVariable { name }.into())
+impl From<UnscopedVariable> for Expression {
+    fn from(variable: UnscopedVariable) -> Expression {
+        Expression::Variable(variable.into())
     }
 }
 
