@@ -5,10 +5,10 @@
 // Please see the LICENSE-APACHE or LICENSE-MIT files in this distribution for license details.
 // ------------------------------------------------------------------------------------------------
 
-use std::collections::HashMap;
-
 use anyhow::Context as ErrorContext;
+use std::collections::HashMap;
 use thiserror::Error;
+use tree_sitter::CaptureQuantifier;
 use tree_sitter::QueryCursor;
 use tree_sitter::QueryMatch;
 use tree_sitter::Tree;
@@ -18,12 +18,15 @@ use crate::ast::AddGraphNodeAttribute;
 use crate::ast::Assign;
 use crate::ast::Call;
 use crate::ast::Capture;
+use crate::ast::Condition;
 use crate::ast::CreateEdge;
 use crate::ast::CreateGraphNode;
 use crate::ast::DeclareImmutable;
 use crate::ast::DeclareMutable;
 use crate::ast::Expression;
 use crate::ast::File;
+use crate::ast::ForIn;
+use crate::ast::If;
 use crate::ast::IntegerConstant;
 use crate::ast::ListComprehension;
 use crate::ast::Print;
@@ -37,14 +40,12 @@ use crate::ast::StringConstant;
 use crate::ast::UnscopedVariable;
 use crate::ast::Variable;
 use crate::functions::Functions;
-use crate::graph::DisplayWithGraph;
 use crate::graph::Graph;
-use crate::graph::GraphNodeRef;
 use crate::graph::SyntaxNodeRef;
 use crate::graph::Value;
-use crate::Context;
-use crate::DisplayWithContext;
-use crate::Identifier;
+pub use crate::variables::Globals;
+use crate::variables::VariableMap;
+use crate::variables::Variables;
 
 impl File {
     /// Executes this graph DSL file against a source file.  You must provide the parsed syntax
@@ -52,34 +53,52 @@ impl File {
     /// provide the set of functions and global variables that are available during execution.
     pub fn execute<'tree>(
         &self,
-        ctx: &Context,
         tree: &'tree Tree,
         source: &'tree str,
         functions: &mut Functions,
-        globals: &mut Variables,
+        globals: &Globals,
     ) -> Result<Graph<'tree>, ExecutionError> {
         let mut graph = Graph::new();
-        let mut locals = Variables::new();
+        self.execute_into(&mut graph, tree, source, functions, globals)?;
+        Ok(graph)
+    }
+
+    /// Executes this graph DSL file against a source file, saving the results into an existing
+    /// `Graph` instance.  You must provide the parsed syntax tree (`tree`) as well as the source
+    /// text that it was parsed from (`source`).  You also provide the set of functions and global
+    /// variables that are available during execution. This variant is useful when you need to
+    /// “pre-seed” the graph with some predefined nodes and/or edges before executing the DSL file.
+    pub fn execute_into<'tree>(
+        &self,
+        graph: &mut Graph<'tree>,
+        tree: &'tree Tree,
+        source: &'tree str,
+        functions: &mut Functions,
+        globals: &Globals,
+    ) -> Result<(), ExecutionError> {
+        if tree.root_node().has_error() {
+            return Err(ExecutionError::ParseTreeHasErrors);
+        }
+        let mut locals = VariableMap::new();
         let mut scoped = ScopedVariables::new();
-        let mut current_regex_matches = Vec::new();
+        let current_regex_captures = Vec::new();
         let mut function_parameters = Vec::new();
         let mut cursor = QueryCursor::new();
         for stanza in &self.stanzas {
             stanza.execute(
-                ctx,
                 tree,
                 source,
-                &mut graph,
+                graph,
                 functions,
                 globals,
                 &mut locals,
                 &mut scoped,
-                &mut current_regex_matches,
+                &current_regex_captures,
                 &mut function_parameters,
                 &mut cursor,
             )?;
         }
-        Ok(graph)
+        Ok(())
     }
 }
 
@@ -88,6 +107,10 @@ impl File {
 pub enum ExecutionError {
     #[error("Cannot assign immutable variable {0}")]
     CannotAssignImmutableVariable(String),
+    #[error("Cannot assign scoped variable {0}")]
+    CannotAssignScopedVariable(String),
+    #[error("Cannot define mutable scoped variable {0}")]
+    CannotDefineMutableScopedVariable(String),
     #[error("Duplicate attribute {0}")]
     DuplicateAttribute(String),
     #[error("Duplicate edge {0}")]
@@ -96,6 +119,10 @@ pub enum ExecutionError {
     DuplicateVariable(String),
     #[error("Expected a graph node reference {0}")]
     ExpectedGraphNode(String),
+    #[error("Expected a list {0}")]
+    ExpectedList(String),
+    #[error("Expected a boolean {0}")]
+    ExpectedBoolean(String),
     #[error("Expected an integer {0}")]
     ExpectedInteger(String),
     #[error("Expected a string {0}")]
@@ -106,16 +133,28 @@ pub enum ExecutionError {
     InvalidParameters(String),
     #[error("Scoped variables can only be attached to syntax nodes {0}")]
     InvalidVariableScope(String),
+    #[error("Recursively defined scoped variable {0}")]
+    RecursivelyDefinedScopedVariable(String),
+    #[error("Recursively defined variable {0}")]
+    RecursivelyDefinedVariable(String),
     #[error("Undefined capture {0}")]
     UndefinedCapture(String),
     #[error("Undefined function {0}")]
     UndefinedFunction(String),
     #[error("Undefined regex capture {0}")]
     UndefinedRegexCapture(String),
+    #[error("Undefined scoped variable {0}")]
+    UndefinedScopedVariable(String),
+    #[error("Empty regex capture {0}")]
+    EmptyRegexCapture(String),
     #[error("Undefined edge {0}")]
     UndefinedEdge(String),
     #[error("Undefined variable {0}")]
     UndefinedVariable(String),
+    #[error("Cannot add scoped variable after being forced {0}")]
+    VariableScopesAlreadyForced(String),
+    #[error("Parse tree has errors")]
+    ParseTreeHasErrors,
     #[error(transparent)]
     Other(#[from] anyhow::Error),
 }
@@ -130,112 +169,45 @@ impl ExecutionError {
     }
 }
 
-struct ExecutionContext<'a, 'tree> {
-    ctx: &'a Context,
+struct ExecutionContext<'a, 'g, 's, 'tree> {
     source: &'tree str,
     graph: &'a mut Graph<'tree>,
     functions: &'a mut Functions,
-    globals: &'a mut Variables,
-    locals: &'a mut Variables,
-    scoped: &'a mut ScopedVariables,
-    current_regex_matches: &'a mut Vec<String>,
+    globals: &'a Globals<'g>,
+    locals: &'a mut dyn Variables<Value>,
+    scoped: &'a mut ScopedVariables<'s>,
+    current_regex_captures: &'a Vec<String>,
     function_parameters: &'a mut Vec<Value>,
     mat: &'a QueryMatch<'a, 'tree>,
 }
 
-/// An environment of named variables
-#[derive(Default)]
-pub struct Variables {
-    values: Vec<NamedVariable>,
+struct ScopedVariables<'a> {
+    scopes: HashMap<SyntaxNodeRef, VariableMap<'a, Value>>,
 }
 
-struct NamedVariable {
-    name: Identifier,
-    value: Value,
-    mutable: bool,
-}
-
-impl Variables {
-    /// Creates a new, empty environment of variables.
-    pub fn new() -> Variables {
-        Variables::default()
-    }
-
-    /// Adds a new mutable variable to an environment, returning an error if the variable already
-    /// exists.
-    pub fn add_mutable(&mut self, name: Identifier, value: Value) -> Result<(), ()> {
-        self.add(name, value, true)
-    }
-
-    /// Adds a new immutable variable to an environment, returning an error if the variable already
-    /// exists.
-    pub fn add_immutable(&mut self, name: Identifier, value: Value) -> Result<(), ()> {
-        self.add(name, value, false)
-    }
-
-    fn add(&mut self, name: Identifier, value: Value, mutable: bool) -> Result<(), ()> {
-        match self.values.binary_search_by_key(&name, |v| v.name) {
-            Ok(_) => Err(()),
-            Err(index) => {
-                let variable = NamedVariable {
-                    name,
-                    value,
-                    mutable,
-                };
-                self.values.insert(index, variable);
-                Ok(())
-            }
+impl<'a> ScopedVariables<'a> {
+    fn new() -> Self {
+        Self {
+            scopes: HashMap::new(),
         }
     }
 
-    /// Returns the value of a variable, if it exists in this environment.
-    pub fn get(&self, name: Identifier) -> Option<&Value> {
-        self.values
-            .binary_search_by_key(&name, |v| v.name)
-            .ok()
-            .map(|index| &self.values[index].value)
-    }
-
-    fn resolve(&mut self, name: Identifier) -> Option<&mut NamedVariable> {
-        self.values
-            .binary_search_by_key(&name, |v| v.name)
-            .ok()
-            .map(move |index| &mut self.values[index])
-    }
-
-    /// Clears this list of variables.
-    pub fn clear(&mut self) {
-        self.values.clear();
-    }
-}
-
-#[derive(Default)]
-struct ScopedVariables {
-    scopes: HashMap<SyntaxNodeRef, Variables>,
-}
-
-impl ScopedVariables {
-    fn new() -> ScopedVariables {
-        ScopedVariables::default()
-    }
-
-    fn get(&mut self, scope: SyntaxNodeRef) -> &mut Variables {
-        self.scopes.entry(scope).or_default()
+    fn get(&mut self, scope: SyntaxNodeRef) -> &mut VariableMap<'a, Value> {
+        self.scopes.entry(scope).or_insert(VariableMap::new())
     }
 }
 
 impl Stanza {
-    fn execute<'tree>(
+    fn execute<'l, 's, 'tree>(
         &self,
-        ctx: &Context,
         tree: &'tree Tree,
         source: &'tree str,
         graph: &mut Graph<'tree>,
         functions: &mut Functions,
-        globals: &mut Variables,
-        locals: &mut Variables,
-        scoped: &mut ScopedVariables,
-        current_regex_matches: &mut Vec<String>,
+        globals: &Globals,
+        locals: &mut VariableMap<'l, Value>,
+        scoped: &mut ScopedVariables<'s>,
+        current_regex_captures: &Vec<String>,
         function_parameters: &mut Vec<Value>,
         cursor: &mut QueryCursor,
     ) -> Result<(), ExecutionError> {
@@ -243,21 +215,20 @@ impl Stanza {
         for mat in matches {
             locals.clear();
             let mut exec = ExecutionContext {
-                ctx,
                 source,
                 graph,
                 functions,
                 globals,
                 locals,
                 scoped,
-                current_regex_matches,
+                current_regex_captures,
                 function_parameters,
                 mat: &mat,
             };
             for statement in &self.statements {
                 statement
                     .execute(&mut exec)
-                    .with_context(|| format!("Executing {}", statement.display_with(exec.ctx)))?;
+                    .with_context(|| format!("Executing {}", statement))?;
             }
         }
         Ok(())
@@ -276,6 +247,8 @@ impl Statement {
             Statement::AddEdgeAttribute(statement) => statement.execute(exec),
             Statement::Scan(statement) => statement.execute(exec),
             Statement::Print(statement) => statement.execute(exec),
+            Statement::If(statement) => statement.execute(exec),
+            Statement::ForIn(statement) => statement.execute(exec),
         }
     }
 }
@@ -297,16 +270,7 @@ impl DeclareMutable {
 impl Assign {
     fn execute(&self, exec: &mut ExecutionContext) -> Result<(), ExecutionError> {
         let value = self.value.evaluate(exec)?;
-        let variable = self.variable.resolve(exec)?;
-        if !variable.mutable {
-            return Err(ExecutionError::CannotAssignImmutableVariable(format!(
-                " {} in {}",
-                self.variable.display_with(exec.ctx),
-                self.display_with(exec.ctx)
-            )));
-        }
-        variable.value = value;
-        Ok(())
+        self.variable.set(exec, value)
     }
 }
 
@@ -320,18 +284,16 @@ impl CreateGraphNode {
 
 impl AddGraphNodeAttribute {
     fn execute(&self, exec: &mut ExecutionContext) -> Result<(), ExecutionError> {
-        let node = self.node.evaluate_as_node(exec)?;
+        let node = self.node.evaluate(exec)?.into_graph_node_ref()?;
         for attribute in &self.attributes {
             let value = attribute.value.evaluate(exec)?;
             exec.graph[node]
                 .attributes
-                .add(attribute.name, value)
+                .add(attribute.name.clone(), value)
                 .map_err(|_| {
                     ExecutionError::DuplicateAttribute(format!(
                         " {} on graph node ({}) in {}",
-                        attribute.name.display_with(exec.ctx),
-                        node.display_with(exec.graph),
-                        self.display_with(exec.ctx),
+                        attribute.name, node, self,
                     ))
                 })?;
         }
@@ -341,15 +303,12 @@ impl AddGraphNodeAttribute {
 
 impl CreateEdge {
     fn execute(&self, exec: &mut ExecutionContext) -> Result<(), ExecutionError> {
-        let source = self.source.evaluate_as_node(exec)?;
-        let sink = self.sink.evaluate_as_node(exec)?;
-        let ctx = exec.ctx;
+        let source = self.source.evaluate(exec)?.into_graph_node_ref()?;
+        let sink = self.sink.evaluate(exec)?.into_graph_node_ref()?;
         if let Err(_) = exec.graph[source].add_edge(sink) {
             Err(ExecutionError::DuplicateEdge(format!(
                 "({} -> {}) in {}",
-                source.display_with(exec.graph),
-                sink.display_with(exec.graph),
-                self.display_with(ctx)
+                source, sink, self,
             )))?;
         }
         Ok(())
@@ -358,28 +317,25 @@ impl CreateEdge {
 
 impl AddEdgeAttribute {
     fn execute(&self, exec: &mut ExecutionContext) -> Result<(), ExecutionError> {
-        let source = self.source.evaluate_as_node(exec)?;
-        let sink = self.sink.evaluate_as_node(exec)?;
+        let source = self.source.evaluate(exec)?.into_graph_node_ref()?;
+        let sink = self.sink.evaluate(exec)?.into_graph_node_ref()?;
         for attribute in &self.attributes {
             let value = attribute.value.evaluate(exec)?;
             let edge = match exec.graph[source].get_edge_mut(sink) {
                 Some(edge) => Ok(edge),
                 None => Err(ExecutionError::UndefinedEdge(format!(
                     "({} -> {}) in {}",
-                    source.display_with(exec.graph),
-                    sink.display_with(exec.graph),
-                    self.display_with(exec.ctx)
+                    source, sink, self,
                 ))),
             }?;
-            edge.attributes.add(attribute.name, value).map_err(|_| {
-                ExecutionError::DuplicateAttribute(format!(
-                    " {} on edge ({} -> {}) in {}",
-                    attribute.name.display_with(exec.ctx),
-                    source.display_with(exec.graph),
-                    sink.display_with(exec.graph),
-                    self.display_with(exec.ctx),
-                ))
-            })?;
+            edge.attributes
+                .add(attribute.name.clone(), value)
+                .map_err(|_| {
+                    ExecutionError::DuplicateAttribute(format!(
+                        " {} on edge ({} -> {}) in {}",
+                        attribute.name, source, sink, self,
+                    ))
+                })?;
         }
         Ok(())
     }
@@ -387,7 +343,7 @@ impl AddEdgeAttribute {
 
 impl Scan {
     fn execute(&self, exec: &mut ExecutionContext) -> Result<(), ExecutionError> {
-        let match_string = self.value.evaluate(exec)?.into_string(exec.graph)?;
+        let match_string = self.value.evaluate(exec)?.into_string()?;
 
         let mut i = 0;
         let mut matches = Vec::new();
@@ -396,6 +352,12 @@ impl Scan {
             for (index, arm) in self.arms.iter().enumerate() {
                 let captures = arm.regex.captures(&match_string[i..]);
                 if let Some(captures) = captures {
+                    if captures.get(0).unwrap().range().is_empty() {
+                        return Err(ExecutionError::EmptyRegexCapture(format!(
+                            "for regular expression /{}/",
+                            arm.regex
+                        )));
+                    }
                     matches.push((captures, index));
                 }
             }
@@ -411,15 +373,30 @@ impl Scan {
 
             let (regex_captures, block_index) = &matches[0];
             let arm = &self.arms[*block_index];
+
+            let mut current_regex_captures = Vec::new();
+            for regex_capture in regex_captures.iter() {
+                current_regex_captures
+                    .push(regex_capture.map(|m| m.as_str()).unwrap_or("").to_string());
+            }
+
+            let mut arm_locals = VariableMap::new_child(exec.locals);
+            let mut arm_exec = ExecutionContext {
+                source: exec.source,
+                graph: exec.graph,
+                functions: exec.functions,
+                globals: exec.globals,
+                locals: &mut arm_locals,
+                scoped: exec.scoped,
+                current_regex_captures: &current_regex_captures,
+                function_parameters: exec.function_parameters,
+                mat: exec.mat,
+            };
+
             for statement in &arm.statements {
-                exec.current_regex_matches.clear();
-                for regex_capture in regex_captures.iter() {
-                    exec.current_regex_matches
-                        .push(regex_capture.map(|m| m.as_str()).unwrap_or("").to_string());
-                }
                 statement
-                    .execute(exec)
-                    .with_context(|| format!("Executing {}", statement.display_with(exec.ctx)))
+                    .execute(&mut arm_exec)
+                    .with_context(|| format!("Executing {}", statement))
                     .with_context(|| {
                         format!(
                             "Matching {} with arm \"{}\" {{ ... }}",
@@ -438,10 +415,80 @@ impl Scan {
 impl Print {
     fn execute(&self, exec: &mut ExecutionContext) -> Result<(), ExecutionError> {
         for value in &self.values {
-            let value = value.evaluate(exec)?;
-            print!("{}", value.display_with(exec.graph));
+            if let Expression::StringConstant(expr) = value {
+                eprint!("{}", expr.value);
+            } else {
+                let value = value.evaluate(exec)?;
+                eprint!("{}", value);
+            }
+        }
+        eprintln!();
+        Ok(())
+    }
+}
+
+impl If {
+    fn execute(&self, exec: &mut ExecutionContext) -> Result<(), ExecutionError> {
+        for arm in &self.arms {
+            let mut result = true;
+            for condition in &arm.conditions {
+                result &= condition.test(exec)?;
+            }
+            if result {
+                let mut arm_locals = VariableMap::new_child(exec.locals);
+                let mut arm_exec = ExecutionContext {
+                    source: exec.source,
+                    graph: exec.graph,
+                    functions: exec.functions,
+                    globals: exec.globals,
+                    locals: &mut arm_locals,
+                    scoped: exec.scoped,
+                    current_regex_captures: exec.current_regex_captures,
+                    function_parameters: exec.function_parameters,
+                    mat: exec.mat,
+                };
+                for stmt in &arm.statements {
+                    stmt.execute(&mut arm_exec)?;
+                }
+                break;
+            }
+        }
+        Ok(())
+    }
+}
+
+impl Condition {
+    fn test(&self, exec: &mut ExecutionContext) -> Result<bool, ExecutionError> {
+        match self {
+            Condition::Some { value, .. } => Ok(!value.evaluate(exec)?.is_null()),
+            Condition::None { value, .. } => Ok(value.evaluate(exec)?.is_null()),
+            Condition::Bool { value, .. } => Ok(value.evaluate(exec)?.into_boolean()?),
+        }
+    }
+}
+
+impl ForIn {
+    fn execute(&self, exec: &mut ExecutionContext) -> Result<(), ExecutionError> {
+        let values = self.value.evaluate(exec)?.into_list()?;
+        let mut loop_locals = VariableMap::new_child(exec.locals);
+        for value in values {
+            loop_locals.clear();
+            let mut loop_exec = ExecutionContext {
+                source: exec.source,
+                graph: exec.graph,
+                functions: exec.functions,
+                globals: exec.globals,
+                locals: &mut loop_locals,
+                scoped: exec.scoped,
+                current_regex_captures: exec.current_regex_captures,
+                function_parameters: exec.function_parameters,
+                mat: exec.mat,
+            };
+            self.variable.add(&mut loop_exec, value, false)?;
+            for stmt in &self.statements {
+                stmt.execute(&mut loop_exec)?;
+            }
         }
-        println!("");
         Ok(())
     }
 }
@@ -462,21 +509,6 @@ impl Expression {
             Expression::RegexCapture(expr) => expr.evaluate(exec),
         }
     }
-
-    fn evaluate_as_node(
-        &self,
-        exec: &mut ExecutionContext,
-    ) -> Result<GraphNodeRef, ExecutionError> {
-        let node = self.evaluate(exec)?;
-        match node {
-            Value::GraphNode(node) => Ok(node),
-            _ => Err(ExecutionError::ExpectedGraphNode(format!(
-                " {}, got {}",
-                self.display_with(exec.ctx),
-                node.display_with(exec.graph)
-            ))),
-        }
-    }
 }
 
 impl IntegerConstant {
@@ -515,60 +547,54 @@ impl SetComprehension {
 
 impl Capture {
     fn evaluate(&self, exec: &mut ExecutionContext) -> Result<Value, ExecutionError> {
-        for capture in exec.mat.captures {
-            if capture.index as usize == self.index {
-                let syntax_node = exec.graph.add_syntax_node(capture.node);
-                return Ok(Value::SyntaxNode(syntax_node));
-            }
-        }
-        Err(ExecutionError::UndefinedCapture(format!(
-            "{}",
-            self.display_with(exec.ctx)
-        )))
+        Ok(query_capture_value(
+            self.stanza_capture_index,
+            self.quantifier,
+            exec.mat,
+            exec.graph,
+        ))
     }
 }
 
 impl Call {
     fn evaluate(&self, exec: &mut ExecutionContext) -> Result<Value, ExecutionError> {
-        exec.function_parameters.clear();
         for parameter in &self.parameters {
             let parameter = parameter.evaluate(exec)?;
             exec.function_parameters.push(parameter);
         }
         exec.functions.call(
-            exec.ctx,
-            self.function,
+            &self.function,
             exec.graph,
             exec.source,
-            &mut exec.function_parameters.drain(..),
+            &mut exec
+                .function_parameters
+                .drain(exec.function_parameters.len() - self.parameters.len()..),
         )
     }
 }
 
 impl RegexCapture {
     fn evaluate(&self, exec: &mut ExecutionContext) -> Result<Value, ExecutionError> {
-        let capture = exec.current_regex_matches.get(self.match_index).ok_or(
-            ExecutionError::UndefinedRegexCapture(format!("{}", self.display_with(exec.ctx))),
-        )?;
+        let capture = exec
+            .current_regex_captures
+            .get(self.match_index)
+            .ok_or(ExecutionError::UndefinedRegexCapture(format!("{}", self)))?;
         Ok(Value::String(capture.clone()))
     }
 }
 
 impl Variable {
     fn evaluate(&self, exec: &mut ExecutionContext) -> Result<Value, ExecutionError> {
-        let variable = self.resolve(exec)?;
-        Ok(variable.value.clone())
+        let value = self.get(exec)?;
+        Ok(value.clone())
     }
 }
 
 impl Variable {
-    fn resolve<'a>(
-        &self,
-        exec: &'a mut ExecutionContext,
-    ) -> Result<&'a mut NamedVariable, ExecutionError> {
+    fn get<'a>(&self, exec: &'a mut ExecutionContext) -> Result<&'a Value, ExecutionError> {
         match self {
-            Variable::Scoped(variable) => variable.resolve(exec),
-            Variable::Unscoped(variable) => variable.resolve(exec),
+            Variable::Scoped(variable) => variable.get(exec),
+            Variable::Unscoped(variable) => variable.get(exec),
         }
     }
 
@@ -583,71 +609,86 @@ impl Variable {
             Variable::Unscoped(variable) => variable.add(exec, value, mutable),
         }
     }
+
+    fn set(&self, exec: &mut ExecutionContext, value: Value) -> Result<(), ExecutionError> {
+        match self {
+            Variable::Scoped(variable) => variable.set(exec, value),
+            Variable::Unscoped(variable) => variable.set(exec, value),
+        }
+    }
 }
 
 impl ScopedVariable {
-    fn resolve<'a>(
-        &self,
-        exec: &'a mut ExecutionContext,
-    ) -> Result<&'a mut NamedVariable, ExecutionError> {
+    fn get<'a>(&self, exec: &'a mut ExecutionContext) -> Result<&'a Value, ExecutionError> {
         let scope = self.scope.evaluate(exec)?;
         let scope = match scope {
             Value::SyntaxNode(scope) => scope,
             _ => {
                 return Err(ExecutionError::InvalidVariableScope(format!(
                     " got {}",
-                    scope.display_with(exec.graph)
+                    scope
+                )))
+            }
+        };
+        let variables = exec.scoped.get(scope);
+        if let Some(value) = variables.get(&self.name) {
+            Ok(value)
+        } else {
+            Err(ExecutionError::UndefinedVariable(format!(
+                "{} on node {}",
+                self, scope,
+            )))
+        }
+    }
+
+    fn add(
+        &self,
+        exec: &mut ExecutionContext,
+        value: Value,
+        mutable: bool,
+    ) -> Result<(), ExecutionError> {
+        let scope = self.scope.evaluate(exec)?;
+        let scope = match scope {
+            Value::SyntaxNode(scope) => scope,
+            _ => {
+                return Err(ExecutionError::InvalidVariableScope(format!(
+                    " got {}",
+                    scope
                 )))
             }
         };
         let variables = exec.scoped.get(scope);
         variables
-            .resolve(self.name)
-            .ok_or(ExecutionError::UndefinedVariable(format!(
-                "{} on node {}",
-                self.display_with(exec.ctx),
-                scope.display_with(exec.graph)
-            )))
+            .add(self.name.clone(), value, mutable)
+            .map_err(|_| ExecutionError::DuplicateVariable(format!(" {}", self)))
     }
 
-    fn add(
-        &self,
-        exec: &mut ExecutionContext,
-        value: Value,
-        mutable: bool,
-    ) -> Result<(), ExecutionError> {
+    fn set(&self, exec: &mut ExecutionContext, value: Value) -> Result<(), ExecutionError> {
         let scope = self.scope.evaluate(exec)?;
         let scope = match scope {
             Value::SyntaxNode(scope) => scope,
             _ => {
                 return Err(ExecutionError::InvalidVariableScope(format!(
                     " got {}",
-                    scope.display_with(exec.graph)
+                    scope,
                 )))
             }
         };
         let variables = exec.scoped.get(scope);
-        variables.add(self.name, value, mutable).map_err(|_| {
-            ExecutionError::DuplicateVariable(format!(" {}", self.display_with(exec.ctx)))
-        })
+        variables
+            .set(self.name.clone(), value)
+            .map_err(|_| ExecutionError::DuplicateVariable(format!(" {}", self)))
     }
 }
 
 impl UnscopedVariable {
-    fn resolve<'a>(
-        &self,
-        exec: &'a mut ExecutionContext,
-    ) -> Result<&'a mut NamedVariable, ExecutionError> {
-        if let Some(variable) = exec.globals.resolve(self.name) {
-            return Ok(variable);
+    fn get<'a>(&self, exec: &'a mut ExecutionContext) -> Result<&'a Value, ExecutionError> {
+        if let Some(value) = exec.globals.get(&self.name) {
+            Some(value)
+        } else {
+            exec.locals.get(&self.name)
         }
-        if let Some(variable) = exec.locals.resolve(self.name) {
-            return Ok(variable);
-        }
-        Err(ExecutionError::UndefinedVariable(format!(
-            "{}",
-            self.display_with(exec.ctx)
-        )))
+        .ok_or_else(|| ExecutionError::UndefinedVariable(format!("{}", self)))
     }
 
     fn add(
@@ -656,14 +697,64 @@ impl UnscopedVariable {
         value: Value,
         mutable: bool,
     ) -> Result<(), ExecutionError> {
-        if exec.globals.get(self.name).is_some() {
+        if exec.globals.get(&self.name).is_some() {
             return Err(ExecutionError::DuplicateVariable(format!(
                 " global {}",
-                self.display_with(exec.ctx)
+                self,
             )));
         }
-        exec.locals.add(self.name, value, mutable).map_err(|_| {
-            ExecutionError::DuplicateVariable(format!(" local {}", self.display_with(exec.ctx)))
+        exec.locals
+            .add(self.name.clone(), value, mutable)
+            .map_err(|_| ExecutionError::DuplicateVariable(format!(" local {}", self)))
+    }
+
+    fn set(&self, exec: &mut ExecutionContext, value: Value) -> Result<(), ExecutionError> {
+        if exec.globals.get(&self.name).is_some() {
+            return Err(ExecutionError::CannotAssignImmutableVariable(format!(
+                " global {}",
+                self,
+            )));
+        }
+        exec.locals.set(self.name.clone(), value).map_err(|_| {
+            if exec.locals.get(&self.name).is_some() {
+                ExecutionError::CannotAssignImmutableVariable(format!("{}", self))
+            } else {
+                ExecutionError::UndefinedVariable(format!("{}", self))
+            }
         })
     }
 }
+
+/// Get the value for the given capture, considering the suffix
+pub(crate) fn query_capture_value<'tree>(
+    index: usize,
+    quantifier: CaptureQuantifier,
+    mat: &QueryMatch<'_, 'tree>,
+    graph: &mut Graph<'tree>,
+) -> Value {
+    let mut nodes = mat
+        .captures
+        .iter()
+        .filter(|c| c.index as usize == index)
+        .map(|c| c.node);
+    match quantifier {
+        CaptureQuantifier::Zero => panic!("Capture with quantifier 0 has no value"),
+        CaptureQuantifier::One => {
+            let syntax_node = graph.add_syntax_node(nodes.next().unwrap());
+            syntax_node.into()
+        }
+        CaptureQuantifier::ZeroOrMore | CaptureQuantifier::OneOrMore => {
+            let syntax_nodes = nodes
+                .map(|n| graph.add_syntax_node(n.clone()).into())
+                .collect::<Vec<Value>>();
+            syntax_nodes.into()
+        }
+        CaptureQuantifier::ZeroOrOne => match nodes.next() {
+            None => Value::Null.into(),
+            Some(node) => {
+                let syntax_node = graph.add_syntax_node(node);
+                syntax_node.into()
+            }
+        },
+    }
+}
