@@ -11,18 +11,32 @@ use std::str::Chars;
 
 use regex::Regex;
 use thiserror::Error;
+use tree_sitter::CaptureQuantifier::Zero;
 use tree_sitter::Language;
 use tree_sitter::Query;
 use tree_sitter::QueryError;
 
 use crate::ast;
-use crate::Context;
 use crate::Identifier;
 
+pub const FULL_MATCH: &str = "__tsg__full_match";
+
 impl ast::File {
+    /// Parses a graph DSL file, returning a new `File` instance.
+    pub fn from_str(language: Language, source: &str) -> Result<Self, ParseError> {
+        let mut file = ast::File::new(language);
+        #[allow(deprecated)]
+        file.parse(source)?;
+        file.check()?;
+        Ok(file)
+    }
+
     /// Parses a graph DSL file, adding its content to an existing `File` instance.
-    pub fn parse(&mut self, ctx: &mut Context, content: &str) -> Result<(), ParseError> {
-        Parser::new(ctx, content).parse_into_file(self)
+    #[deprecated(
+        note = "Parsing multiple times into the same `File` instance is unsound. Use `File::from_str` instead."
+    )]
+    pub fn parse(&mut self, content: &str) -> Result<(), ParseError> {
+        Parser::new(content).parse_into_file(self)
     }
 }
 
@@ -33,6 +47,8 @@ pub enum ParseError {
     ExpectedToken(&'static str, Location),
     #[error("Expected variable name at {0}")]
     ExpectedVariable(Location),
+    #[error("Expected unscoped variable at {0}")]
+    ExpectedUnscopedVariable(Location),
     #[error("Invalid regular expression /{0}/ at {1}")]
     InvalidRegex(String, Location),
     #[error("Expected integer constant in regex capture at {0}")]
@@ -41,8 +57,6 @@ pub enum ParseError {
     // row/column from the start of the query, not from the start of the file.
     #[error("Invalid query pattern: {}", _0.message)]
     QueryError(#[from] QueryError),
-    #[error("Undefined query capture '{0}' at {1}")]
-    UndefinedCapture(String, Location),
     #[error("Unexpected character '{0}' in {1} at {2}")]
     UnexpectedCharacter(char, &'static str, Location),
     #[error("Unexpected end of file at {0}")]
@@ -51,6 +65,20 @@ pub enum ParseError {
     UnexpectedKeyword(String, Location),
     #[error("Unexpected literal '#{0}' at {1}")]
     UnexpectedLiteral(String, Location),
+    #[error(transparent)]
+    Check(#[from] crate::checker::CheckError),
+    #[error(transparent)]
+    Other(#[from] anyhow::Error),
+}
+
+impl ParseError {
+    /// Wraps an existing [`std::error::Error`][] as an execution error
+    pub fn other<E>(err: E) -> ParseError
+    where
+        E: Into<anyhow::Error>,
+    {
+        Self::Other(err.into())
+    }
 }
 
 /// The location of a graph DSL entity within its file
@@ -78,23 +106,11 @@ impl Display for Location {
 }
 
 struct Parser<'a> {
-    ctx: &'a mut Context,
     source: &'a str,
     chars: Peekable<Chars<'a>>,
     offset: usize,
     location: Location,
-    // Keywords
-    attr_keyword: Identifier,
-    edge_keyword: Identifier,
-    false_keyword: Identifier,
-    let_keyword: Identifier,
-    node_keyword: Identifier,
-    null_keyword: Identifier,
-    print_keyword: Identifier,
-    scan_keyword: Identifier,
-    set_keyword: Identifier,
-    true_keyword: Identifier,
-    var_keyword: Identifier,
+    query_source: String,
 }
 
 fn is_ident_start(c: char) -> bool {
@@ -106,41 +122,20 @@ fn is_ident(c: char) -> bool {
 }
 
 impl<'a> Parser<'a> {
-    fn new(ctx: &'a mut Context, source: &'a str) -> Parser<'a> {
+    fn new(source: &'a str) -> Parser<'a> {
         let chars = source.chars().peekable();
-        let attr_keyword = ctx.add_identifier("attr");
-        let edge_keyword = ctx.add_identifier("edge");
-        let false_keyword = ctx.add_identifier("false");
-        let let_keyword = ctx.add_identifier("let");
-        let node_keyword = ctx.add_identifier("node");
-        let null_keyword = ctx.add_identifier("null");
-        let print_keyword = ctx.add_identifier("print");
-        let scan_keyword = ctx.add_identifier("scan");
-        let set_keyword = ctx.add_identifier("set");
-        let true_keyword = ctx.add_identifier("true");
-        let var_keyword = ctx.add_identifier("var");
+        let query_source = String::with_capacity(source.len());
         Parser {
-            ctx,
             source,
             chars,
             offset: 0,
             location: Location::default(),
-            attr_keyword,
-            edge_keyword,
-            false_keyword,
-            let_keyword,
-            node_keyword,
-            null_keyword,
-            print_keyword,
-            scan_keyword,
-            set_keyword,
-            true_keyword,
-            var_keyword,
+            query_source,
         }
     }
 }
 
-impl Parser<'_> {
+impl<'a> Parser<'a> {
     fn peek(&mut self) -> Result<char, ParseError> {
         self.chars
             .peek()
@@ -162,10 +157,6 @@ impl Parser<'_> {
         Ok(ch)
     }
 
-    fn try_next(&mut self) -> Option<char> {
-        self.next().ok()
-    }
-
     fn skip(&mut self) -> Result<(), ParseError> {
         self.next().map(|_| ())
     }
@@ -219,6 +210,7 @@ impl Parser<'_> {
             file.stanzas.push(stanza);
             self.consume_whitespace();
         }
+        file.query = Some(Query::new(file.language, &self.query_source)?);
         Ok(())
     }
 
@@ -226,10 +218,11 @@ impl Parser<'_> {
         let location = self.location;
         let query = self.parse_query(language)?;
         self.consume_whitespace();
-        let statements = self.parse_statements(&query)?;
+        let statements = self.parse_statements()?;
         Ok(ast::Stanza {
             query,
             statements,
+            full_match_file_capture_index: usize::MAX, // set in checker
             location,
         })
     }
@@ -238,8 +231,13 @@ impl Parser<'_> {
         let query_start = self.offset;
         self.skip_query()?;
         let query_end = self.offset;
-        let query_source = &self.source[query_start..query_end];
-        Ok(Query::new(language, query_source)?)
+        let query_source = self.source[query_start..query_end].to_owned() + "@" + FULL_MATCH;
+        // If tree-sitter allowed us to incrementally add patterns to a query, we wouldn't need
+        // the global query_source and could compute the cpature indices in the AST instead of
+        // having to resolve the capture names at execution time.
+        self.query_source += &query_source;
+        self.query_source += "\n";
+        Ok(Query::new(language, &query_source)?)
     }
 
     fn skip_query(&mut self) -> Result<(), ParseError> {
@@ -283,15 +281,12 @@ impl Parser<'_> {
         }
     }
 
-    fn parse_statements(
-        &mut self,
-        current_query: &Query,
-    ) -> Result<Vec<ast::Statement>, ParseError> {
+    fn parse_statements(&mut self) -> Result<Vec<ast::Statement>, ParseError> {
         self.consume_token("{")?;
         let mut statements = Vec::new();
         self.consume_whitespace();
         while self.peek()? != '}' {
-            let statement = self.parse_statement(current_query)?;
+            let statement = self.parse_statement()?;
             statements.push(statement);
             self.consume_whitespace();
         }
@@ -299,80 +294,91 @@ impl Parser<'_> {
         Ok(statements)
     }
 
-    fn parse_statement(&mut self, current_query: &Query) -> Result<ast::Statement, ParseError> {
+    fn parse_name(&mut self, within: &'static str) -> Result<&'a str, ParseError> {
+        let start = self.offset;
+        let ch = self.next()?;
+        if !is_ident_start(ch) {
+            return Err(ParseError::UnexpectedCharacter(ch, within, self.location));
+        }
+        self.consume_while(is_ident);
+        let end = self.offset;
+        Ok(&self.source[start..end])
+    }
+
+    fn parse_statement(&mut self) -> Result<ast::Statement, ParseError> {
         let keyword_location = self.location;
-        let keyword = self.parse_identifier("keyword")?;
+        let keyword = self.parse_name("keyword")?;
         self.consume_whitespace();
-        if keyword == self.let_keyword {
-            let variable = self.parse_variable(current_query)?;
+        if keyword == "let" {
+            let variable = self.parse_variable()?;
             self.consume_whitespace();
             self.consume_token("=")?;
             self.consume_whitespace();
-            let value = self.parse_expression(current_query)?;
+            let value = self.parse_expression()?;
             Ok(ast::DeclareImmutable {
                 variable,
                 value,
                 location: keyword_location,
             }
             .into())
-        } else if keyword == self.var_keyword {
-            let variable = self.parse_variable(current_query)?;
+        } else if keyword == "var" {
+            let variable = self.parse_variable()?;
             self.consume_whitespace();
             self.consume_token("=")?;
             self.consume_whitespace();
-            let value = self.parse_expression(current_query)?;
+            let value = self.parse_expression()?;
             Ok(ast::DeclareMutable {
                 variable,
                 value,
                 location: keyword_location,
             }
             .into())
-        } else if keyword == self.set_keyword {
-            let variable = self.parse_variable(current_query)?;
+        } else if keyword == "set" {
+            let variable = self.parse_variable()?;
             self.consume_whitespace();
             self.consume_token("=")?;
             self.consume_whitespace();
-            let value = self.parse_expression(current_query)?;
+            let value = self.parse_expression()?;
             Ok(ast::Assign {
                 variable,
                 value,
                 location: keyword_location,
             }
             .into())
-        } else if keyword == self.node_keyword {
-            let node = self.parse_variable(current_query)?;
+        } else if keyword == "node" {
+            let node = self.parse_variable()?;
             Ok(ast::CreateGraphNode {
                 node,
                 location: keyword_location,
             }
             .into())
-        } else if keyword == self.edge_keyword {
-            let source = self.parse_expression(current_query)?;
+        } else if keyword == "edge" {
+            let source = self.parse_expression()?;
             self.consume_whitespace();
             self.consume_token("->")?;
             self.consume_whitespace();
-            let sink = self.parse_expression(current_query)?;
+            let sink = self.parse_expression()?;
             Ok(ast::CreateEdge {
                 source,
                 sink,
                 location: keyword_location,
             }
             .into())
-        } else if keyword == self.attr_keyword {
+        } else if keyword == "attr" {
             self.consume_token("(")?;
             self.consume_whitespace();
-            let node_or_source = self.parse_expression(current_query)?;
+            let node_or_source = self.parse_expression()?;
             self.consume_whitespace();
 
             if self.peek()? == '-' {
                 let source = node_or_source;
                 self.consume_token("->")?;
                 self.consume_whitespace();
-                let sink = self.parse_expression(current_query)?;
+                let sink = self.parse_expression()?;
                 self.consume_whitespace();
                 self.consume_token(")")?;
                 self.consume_whitespace();
-                let attributes = self.parse_attributes(current_query)?;
+                let attributes = self.parse_attributes()?;
                 Ok(ast::AddEdgeAttribute {
                     source,
                     sink,
@@ -385,7 +391,7 @@ impl Parser<'_> {
                 self.consume_whitespace();
                 self.consume_token(")")?;
                 self.consume_whitespace();
-                let attributes = self.parse_attributes(current_query)?;
+                let attributes = self.parse_attributes()?;
                 Ok(ast::AddGraphNodeAttribute {
                     node,
                     attributes,
@@ -393,20 +399,23 @@ impl Parser<'_> {
                 }
                 .into())
             }
-        } else if keyword == self.print_keyword {
-            let mut values = vec![self.parse_expression(current_query)?];
+        } else if keyword == "print" {
+            let mut values = vec![self.parse_expression()?];
             self.consume_whitespace();
-            while self.try_next() == Some(',') {
-                values.push(self.parse_expression(current_query)?);
+            while self.try_peek() == Some(',') {
+                self.consume_token(",")?;
+                self.consume_whitespace();
+                values.push(self.parse_expression()?);
                 self.consume_whitespace();
             }
+            self.consume_whitespace();
             Ok(ast::Print {
                 values,
                 location: keyword_location,
             }
             .into())
-        } else if keyword == self.scan_keyword {
-            let value = self.parse_expression(current_query)?;
+        } else if keyword == "scan" {
+            let value = self.parse_expression()?;
             self.consume_whitespace();
             self.consume_token("{")?;
             self.consume_whitespace();
@@ -414,14 +423,10 @@ impl Parser<'_> {
             while self.peek()? != '}' {
                 let pattern_location = self.location;
                 let pattern = self.parse_string()?;
-                let regex = match Regex::new(&pattern) {
-                    Ok(regex) => regex,
-                    Err(_) => {
-                        return Err(ParseError::InvalidRegex(pattern.into(), pattern_location))
-                    }
-                };
+                let regex = Regex::new(&pattern)
+                    .map_err(|_| ParseError::InvalidRegex(pattern.into(), pattern_location))?;
                 self.consume_whitespace();
-                let statements = self.parse_statements(current_query)?;
+                let statements = self.parse_statements()?;
                 arms.push(ast::ScanArm {
                     regex,
                     statements,
@@ -436,23 +441,131 @@ impl Parser<'_> {
                 location: keyword_location,
             }
             .into())
+        } else if keyword == "if" {
+            let mut arms = Vec::new();
+
+            // if
+            let location = keyword_location;
+            self.consume_whitespace();
+            let conditions = self.parse_conditions()?;
+            self.consume_whitespace();
+            let statements = self.parse_statements()?;
+            self.consume_whitespace();
+            arms.push(ast::IfArm {
+                conditions,
+                statements,
+                location,
+            });
+
+            // elif
+            let mut location = self.location;
+            while let Ok(_) = self.consume_token("elif") {
+                self.consume_whitespace();
+                let conditions = self.parse_conditions()?;
+                self.consume_whitespace();
+                let statements = self.parse_statements()?;
+                self.consume_whitespace();
+                arms.push(ast::IfArm {
+                    conditions,
+                    statements,
+                    location,
+                });
+                self.consume_whitespace();
+                location = self.location;
+            }
+
+            // else
+            let location = self.location;
+            if let Ok(_) = self.consume_token("else") {
+                let conditions = vec![];
+                self.consume_whitespace();
+                let statements = self.parse_statements()?;
+                self.consume_whitespace();
+                arms.push(ast::IfArm {
+                    conditions,
+                    statements,
+                    location,
+                });
+                self.consume_whitespace();
+            }
+
+            Ok(ast::If {
+                arms,
+                location: keyword_location,
+            }
+            .into())
+        } else if keyword == "for" {
+            self.consume_whitespace();
+            let variable = match self.parse_variable()? {
+                ast::Variable::Unscoped(variable) => Ok(variable),
+                ast::Variable::Scoped(variable) => {
+                    Err(ParseError::ExpectedUnscopedVariable(variable.location))
+                }
+            }?;
+            self.consume_whitespace();
+            self.consume_token("in")?;
+            self.consume_whitespace();
+            let value = self.parse_expression()?;
+            self.consume_whitespace();
+            let statements = self.parse_statements()?;
+            Ok(ast::ForIn {
+                variable,
+                value,
+                statements,
+                location: keyword_location,
+            }
+            .into())
         } else {
             Err(ParseError::UnexpectedKeyword(
-                self.ctx.resolve(keyword).into(),
+                keyword.into(),
                 keyword_location,
             ))
         }
     }
 
-    fn parse_identifier(&mut self, within: &'static str) -> Result<Identifier, ParseError> {
-        let start = self.offset;
-        let ch = self.next()?;
-        if !is_ident_start(ch) {
-            return Err(ParseError::UnexpectedCharacter(ch, within, self.location));
+    fn parse_conditions(&mut self) -> Result<Vec<ast::Condition>, ParseError> {
+        let mut conditions = Vec::new();
+        let mut has_next = true;
+        while has_next {
+            conditions.push(self.parse_condition()?);
+            self.consume_whitespace();
+            if let Some(',') = self.try_peek() {
+                self.consume_token(",")?;
+                self.consume_whitespace();
+                has_next = true;
+            } else {
+                has_next = false;
+            }
         }
-        self.consume_while(is_ident);
-        let end = self.offset;
-        Ok(self.ctx.add_identifier(&self.source[start..end]))
+        Ok(conditions)
+    }
+
+    fn parse_condition(&mut self) -> Result<ast::Condition, ParseError> {
+        let location = self.location;
+        let condition = if let Ok(_) = self.consume_token("some") {
+            self.consume_whitespace();
+            let value = self.parse_expression()?;
+            ast::Condition::Some { value, location }
+        } else if let Ok(_) = self.consume_token("none") {
+            self.consume_whitespace();
+            let value = self.parse_expression()?;
+            ast::Condition::None { value, location }
+        } else if let Ok(value) = self.parse_expression() {
+            self.consume_whitespace();
+            ast::Condition::Bool { value, location }
+        } else {
+            return Err(ParseError::ExpectedToken(
+                "(some|none)? EXPRESSION",
+                location,
+            ));
+        };
+        self.consume_whitespace();
+        Ok(condition)
+    }
+
+    fn parse_identifier(&mut self, within: &'static str) -> Result<Identifier, ParseError> {
+        let content = self.parse_name(within)?;
+        Ok(Identifier::from(content))
     }
 
     fn parse_string(&mut self) -> Result<String, ParseError> {
@@ -480,15 +593,21 @@ impl Parser<'_> {
         }
     }
 
-    fn parse_expression(&mut self, current_query: &Query) -> Result<ast::Expression, ParseError> {
+    fn parse_expression(&mut self) -> Result<ast::Expression, ParseError> {
         let mut expression = match self.peek()? {
             '#' => self.parse_literal()?,
             '"' => self.parse_string()?.into(),
-            '@' => self.parse_capture(current_query)?,
-            '$' => self.parse_regex_capture()?,
-            '(' => self.parse_call(current_query)?,
+            '@' => self.parse_capture()?.into(),
+            '$' => self.parse_regex_capture()?.into(),
+            '(' => self.parse_call()?,
+            '[' => self.parse_list()?,
+            '{' => self.parse_set()?,
             ch if ch.is_ascii_digit() => self.parse_integer_constant()?,
-            ch if is_ident_start(ch) => self.parse_identifier("variable name")?.into(),
+            ch if is_ident_start(ch) => {
+                let location = self.location;
+                let name = self.parse_identifier("variable name")?;
+                ast::UnscopedVariable { name, location }.into()
+            }
             ch => {
                 return Err(ParseError::UnexpectedCharacter(
                     ch,
@@ -501,22 +620,28 @@ impl Parser<'_> {
         while self.try_peek() == Some('.') {
             self.skip().unwrap();
             self.consume_whitespace();
+            let location = self.location;
             let scope = Box::new(expression);
             let name = self.parse_identifier("scoped variable name")?;
             self.consume_whitespace();
-            expression = ast::ScopedVariable { scope, name }.into();
+            expression = ast::ScopedVariable {
+                scope,
+                name,
+                location,
+            }
+            .into();
         }
         Ok(expression)
     }
 
-    fn parse_call(&mut self, current_query: &Query) -> Result<ast::Expression, ParseError> {
+    fn parse_call(&mut self) -> Result<ast::Expression, ParseError> {
         self.consume_token("(")?;
         self.consume_whitespace();
         let function = self.parse_identifier("function name")?;
         self.consume_whitespace();
         let mut parameters = Vec::new();
         while self.peek()? != ')' {
-            parameters.push(self.parse_expression(current_query)?);
+            parameters.push(self.parse_expression()?);
             self.consume_whitespace();
         }
         self.consume_token(")")?;
@@ -527,8 +652,37 @@ impl Parser<'_> {
         .into())
     }
 
-    fn parse_capture(&mut self, current_query: &Query) -> Result<ast::Expression, ParseError> {
-        let capture_location = self.location;
+    fn parse_sequence(&mut self, end_marker: char) -> Result<Vec<ast::Expression>, ParseError> {
+        let mut elements = Vec::new();
+        while self.peek()? != end_marker {
+            elements.push(self.parse_expression()?);
+            self.consume_whitespace();
+            if self.peek()? != end_marker {
+                self.consume_token(",")?;
+                self.consume_whitespace();
+            }
+        }
+        Ok(elements)
+    }
+
+    fn parse_list(&mut self) -> Result<ast::Expression, ParseError> {
+        self.consume_token("[")?;
+        self.consume_whitespace();
+        let elements = self.parse_sequence(']')?;
+        self.consume_token("]")?;
+        Ok(ast::ListComprehension { elements }.into())
+    }
+
+    fn parse_set(&mut self) -> Result<ast::Expression, ParseError> {
+        self.consume_token("{")?;
+        self.consume_whitespace();
+        let elements = self.parse_sequence('}')?;
+        self.consume_token("}")?;
+        Ok(ast::SetComprehension { elements }.into())
+    }
+
+    fn parse_capture(&mut self) -> Result<ast::Capture, ParseError> {
+        let location = self.location;
         let start = self.offset;
         self.consume_token("@")?;
         let ch = self.next()?;
@@ -541,22 +695,15 @@ impl Parser<'_> {
         }
         self.consume_while(is_ident);
         let end = self.offset;
-        let capture_name = &self.source[start + 1..end];
-        let index = current_query
-            .capture_names()
-            .iter()
-            .position(|c| c == capture_name);
-        let index = match index {
-            Some(index) => index,
-            None => {
-                return Err(ParseError::UndefinedCapture(
-                    capture_name.into(),
-                    capture_location,
-                ))
-            }
-        };
-        let name = self.ctx.add_identifier(&self.source[start..end]);
-        Ok(ast::Capture { index, name }.into())
+        let name = Identifier::from(&self.source[start + 1..end]);
+        Ok(ast::Capture {
+            name,
+            quantifier: Zero,                 // set in checker
+            file_capture_index: usize::MAX,   // set in checker
+            stanza_capture_index: usize::MAX, // set in checker
+            location,
+        }
+        .into())
     }
 
     fn parse_integer_constant(&mut self) -> Result<ast::Expression, ParseError> {
@@ -571,22 +718,22 @@ impl Parser<'_> {
     fn parse_literal(&mut self) -> Result<ast::Expression, ParseError> {
         let literal_location = self.location;
         self.consume_token("#")?;
-        let literal = self.parse_identifier("literal")?;
-        if literal == self.false_keyword {
+        let literal = self.parse_name("literal")?;
+        if literal == "false" {
             return Ok(ast::Expression::FalseLiteral);
-        } else if literal == self.null_keyword {
+        } else if literal == "null" {
             return Ok(ast::Expression::NullLiteral);
-        } else if literal == self.true_keyword {
+        } else if literal == "true" {
             return Ok(ast::Expression::TrueLiteral);
         } else {
             Err(ParseError::UnexpectedLiteral(
-                self.ctx.resolve(literal).into(),
+                literal.into(),
                 literal_location,
             ))
         }
     }
 
-    fn parse_regex_capture(&mut self) -> Result<ast::Expression, ParseError> {
+    fn parse_regex_capture(&mut self) -> Result<ast::RegexCapture, ParseError> {
         let regex_capture_location = self.location;
         self.consume_token("$")?;
         let start = self.offset;
@@ -599,37 +746,34 @@ impl Parser<'_> {
         Ok(ast::RegexCapture { match_index }.into())
     }
 
-    fn parse_attributes(
-        &mut self,
-        current_query: &Query,
-    ) -> Result<Vec<ast::Attribute>, ParseError> {
-        let mut attributes = vec![self.parse_attribute(current_query)?];
+    fn parse_attributes(&mut self) -> Result<Vec<ast::Attribute>, ParseError> {
+        let mut attributes = vec![self.parse_attribute()?];
         self.consume_whitespace();
         while self.try_peek() == Some(',') {
             self.skip().unwrap();
             self.consume_whitespace();
-            attributes.push(self.parse_attribute(current_query)?);
+            attributes.push(self.parse_attribute()?);
             self.consume_whitespace();
         }
         Ok(attributes)
     }
 
-    fn parse_attribute(&mut self, current_query: &Query) -> Result<ast::Attribute, ParseError> {
+    fn parse_attribute(&mut self) -> Result<ast::Attribute, ParseError> {
         let name = self.parse_identifier("attribute name")?;
         self.consume_whitespace();
         let value = if self.try_peek() == Some('=') {
             self.consume_token("=")?;
             self.consume_whitespace();
-            self.parse_expression(current_query)?
+            self.parse_expression()?
         } else {
             ast::Expression::TrueLiteral
         };
         Ok(ast::Attribute { name, value })
     }
 
-    fn parse_variable(&mut self, current_query: &Query) -> Result<ast::Variable, ParseError> {
+    fn parse_variable(&mut self) -> Result<ast::Variable, ParseError> {
         let expression_location = self.location;
-        match self.parse_expression(current_query)? {
+        match self.parse_expression()? {
             ast::Expression::Variable(variable) => Ok(variable),
             _ => Err(ParseError::ExpectedVariable(expression_location)),
         }
